(   Title:  A utiility for testing bstack effects
     File:  btester.fs
   Author:  David N. Williams
  License:  Public Domain, John Hayes
  Version:  1.4.3, January 18, 2021
Generated:  January 18, 2021 from tester.tmpl

For any code derived from John Hayes' tester program:
)
\ (C) 1995 JOHNS HOPKINS UNIVERSITY / APPLIED PHYSICS LABORATORY
\ MAY BE DISTRIBUTED FREELY AS LONG AS THIS COPYRIGHT NOTICE REMAINS.
(
This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

The rest is public domain, based on the ttester derivative by Anton Ertl
and its xftester derivative by David N. Williams.
)

\ *** USAGE
(
PUBLIC INTERFACE WORDS

   tests:  B{  B->  }B
  config:  MAX-BT-INITIALS  MAX-BT-RESULTS

PUBLIC AUXILIARY WORDS

  errors:  BT-ERROR-XT  BT-ERROR-DEFAULT  BT-ERROR-INDEX
  arrays:  BT-INITIALS  BT-RESULTS  BT-GOALS
           BT-A@  BT-A!  BT-A=

The public interface words are all that are needed for normal testing.

The public auxiliary words are made available for users who want to
implement more elaborate error reporting, such as counting errors, or
dumping results and goals when something is wrong, or reporting a
particular error with the help of BT-ERROR-INDEX.  The file
btester-errors.fs does that, and is loaded by gmpfr-test.fs.

The basic test usage takes the form:
)
\   ( binitials) B{ <bresults> B-> <bgoals> }B ( binitials)
(
Angle brackets around a data designation for a particular stack indicate
code that produces the data.  Here binitials stands for possibly
nonempty bstack contents at the start of the test.  The sequence
starting with B{ saves the binitials and removes them from the bstack,
executes the code <bresults> that produces the bstack bresults, saves
and removes the bresults from the bstack, executes the <bgoals> code
that produces the bstack bgoals, saves and removes them, compares the
saved bresults and bgoals, reports when there is a discrepancy between
the two, and restores the saved binitials.

The words B{, B->, and }B ignore non-bstacks.

Many examples can be found in the file gmpfr-test.fs.

Storage of binitials, bresults, and bgoals includes overflow checking.
The user can change the storage allocation by defining one or both of
the following CONSTANT's *before* loading this file:

  MAX-BT-INITIALS  MAX-BT-RESULTS

The xt in the variable BT-ERROR-XT can be changed only *after* loading
this file.

Loading btester.fs does not change BASE.
)
BASE @ DECIMAL

\ *** ERRORS

VARIABLE BT-ERROR-XT
VARIABLE BT-ERROR-INDEX    \ set to zero after unequal array size
                           \ comparison, or [index+1] after deepest
                           \ unequal element comparison at index

: BT-ERROR-DEFAULT  ( c-addr u -- )
(
Display an error message followed by the line that had the error.
)
  TYPE SOURCE TYPE CR ;
' BT-ERROR-DEFAULT BT-ERROR-XT !

: BT-ERROR BT-ERROR-XT @ EXECUTE ;  \ vectored error reporting

\ *** ARRAYS
(
The first array slot has index zero.  A tick prefix means "address".
)

\ USER CONFIG DEFAULTS
[UNDEFINED] MAX-BT-INITIALS [IF] 32 CONSTANT MAX-BT-INITIALS [THEN]
[UNDEFINED] MAX-BT-RESULTS  [IF] 32 CONSTANT MAX-BT-RESULTS  [THEN]

\ Array buffers begin with a cell-sized count.
CREATE BT-INITIALS MAX-BT-INITIALS 1+ CELLS ALLOT
CREATE BT-RESULTS  MAX-BT-RESULTS  1+ CELLS ALLOT
CREATE BT-GOALS    MAX-BT-RESULTS  1+ CELLS ALLOT

: 0BT-A  ( +n 'buf -- )  \ assume n<>0
(
Initialize the array buffer for garbage collection by clearing the count
and n belement addresses.
)
  swap 1+ cells over + swap DO 0 i ! 1 cells +LOOP ;

MAX-BT-INITIALS BT-INITIALS 0BT-A
MAX-BT-RESULTS  BT-RESULTS  0BT-A
MAX-BT-RESULTS  BT-GOALS    0BT-A

\ Overflow checking to be done elsewhere.
: BT-A!  ( +n 'buf -- ) ( b: b_[n-1] ... b_0 -- )
  2dup ! cell+ swap ( 'elems n) dup
  IF   cells over + swap
       DO i b! 1 cells +LOOP 
  ELSE 2drop THEN ;

: BT-A@  ( 'buf -- +n ) ( b: -- b_[n-1] ... b_0 )
  dup @ dup >r ( 'buf n r: n)
  IF   cell+  r@ 1- cells over + 
       DO i b@ -1 cells +LOOP
  ELSE drop THEN r> ;

[UNDEFINED] CELL- [IF]
  : CELL-  ( addr -- addr-cell )  [ 1 CELLS ] LITERAL - ;
[THEN]

: BT-A=  ( 'buf1 'buf2 -- 0 true | 0 false | index+1 false )
(
Returns:
         0 true:  equal size and elements
        0 false:  sizes unequal
  index+1 false:  sizes equal, deepest unequal elements at index
)
  2dup @ swap @ <>
  IF 2drop 0 false EXIT THEN          \ unequal size
  dup @ 0= IF 2drop 0 true EXIT THEN  \ equal size and empty
  2>r 0 true 2r>
  ( 0 true 'buf1 'buf2)
  dup @ ( count) 1- ( last) >r
  cell+ r@ cells + swap
  cell+ r@ cells + 0 r>
  ( 0 true 'elems2_last 'elems1_last 0 last)
  DO
    2dup b@ b@ b= 0=
    IF  \ unequal elements
      2drop ( 0 true) 2drop i 1+ false 0 0 LEAVE THEN
    cell- swap cell-  \ [i-1]th element addresses interchanged 
  -1 +LOOP
  ( [0 true | i+1 false] x1 x2) 2drop ;

\ *** TESTS

: B{  ( b: binitials -- )
(
Save the binitials for restoration by }B, leaving the bstack empty. 
Clear bresults and bgoals storage.
)
  bdepth dup MAX-BT-INITIALS > ABORT" TOO MANY BINITIALS"
  BT-INITIALS BT-A!   0 BT-RESULTS !   0 BT-GOALS ! ;

: B->  ( b: besults -- )
(
Save the bresults, leaving the bstack empty.
)
  bdepth dup MAX-BT-RESULTS > ABORT" TOO MANY BRESULTS"
  BT-RESULTS BT-A! ;

: }B  ( b: bgoals -- binitials )
(
Save the bgoals, compare them with the saved bresults, and restore the
binitials.  If the number of bresults is not the same as the number of
bgoals, store zero in BT-ERROR-INDEX and execute the BT-ERROR vector. 
Else, if the numbers are the same but there is a deepest comparison
failure at the index-th element, store index+1 in BT-ERROR-INDEX and
execute the BT-ERROR vector. 
)
  bdepth dup MAX-BT-RESULTS > ABORT" TOO MANY BGOALS"
  BT-GOALS BT-A!
  BT-RESULTS BT-GOALS BT-A= ( 0 true | 0 false | index+1 false)
  IF
    ( 0) drop
  ELSE
    dup BT-ERROR-INDEX !
    IF S" INCORRECT BRESULT: " ELSE S" WRONG NUMBER OF BRESULTS: " THEN
    BT-ERROR
  THEN
  BT-INITIALS BT-A@ ( len) drop ;

BASE !
\ *** END of btester.fs
