(   Title:  A utiility for testing mfstack effects
     File:  mftester.fs
   Author:  David N. Williams
  License:  Public Domain, John Hayes
  Version:  1.4.3, January 18, 2021
Generated:  January 18, 2021 from fptester.tmpl

For any code derived from John Hayes' tester program:
)
\ (C) 1995 JOHNS HOPKINS UNIVERSITY / APPLIED PHYSICS LABORATORY
\ MAY BE DISTRIBUTED FREELY AS LONG AS THIS COPYRIGHT NOTICE REMAINS.
(
This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

The rest is public domain, based on the ttester derivative by Anton Ertl
and its xftester derivative by David N. Williams.  Explanatory material
was provided by C. G. Montgomery in March, 2009, with helpful comments
from Krishna Myneni.
)

\ *** USAGE
(
PUBLIC INTERFACE WORDS
 
   tests:  MF{  MF->  }MF
           MFT-ABS-TOLERANCE  MFT-REL-TOLERANCE
           MFT-TEST=-XT
           MFT-DATUM=  MFT-ABS=  MFT-REL=  MFT-REL0=
           SET-MFT-MODE-EXACT  SET-MFT-MODE-ABS  SET-MFT-MODE-REL
           SET-MFT-MODE-REL0
  config:  MAX-MFT-INITIALS  MAX-MFT-RESULTS
 
  PUBLIC AUXILIARY WODS
 
   tests:  MFT-TEST=-XT
           MFT-DATUM=  MFT-ABS=  MFT-REL=  MFT-REL0=
  errors:  MFT-ERROR-XT  MFT-ERROR-DEFAULT  MFT-ERROR-INDEX
  arrays:  MFT-INITIALS  MFT-RESULTS  MFT-GOALS
           MFT-A@  MFT-A!  MFT-A=

The public interface words are all that are needed for normal testing,
including using the built-in fp comparisons and changing their
tolerances.

The public auxiliary words are made available for users who want to add
fp comparisons, or to implement more elaborate error reporting, such as
counting errors, or dumping fp results and goals when something is
wrong, or reporting a particular fp error with the help of
MFT-ERROR-INDEX.  The file mftester-errors.fs does that, and is loaded
by gmpfr-test.fs.

The basic test usage takes the form:
)
\   ( mfinitials) MF{ <mfresults> MF-> <mfgoals> }MF ( mfinitials)
(
Angle brackets around a data designation for a particular stack indicate
code that produces the data.  Here mfinitials stands for possibly
nonempty mfstack contents at the start of the test.  The sequence
starting with MF{ saves the mfinitials and removes them from the
mfstack, executes the code <mfresults> that produces the mfstack
mfresults, saves and removes the mfresults from the mfstack, executes
the <mfgoals> code that produces the mfstack mfgoals, saves and removes
them, compares the saved mfresults and mfgoals, reports when there is a
discrepancy between the two, and restores the saved mfinitials.

The words MF{, MF->, and }MF ignore non-mfstacks.

Many examples can be found in the file gmpfr-test.fs.

Floating-point testing can involve further complications.  There are the
perennial issues of floating-point value comparisons.  Exact equality is
specified by SET-MFT-MODE-EXACT [the default].  If one of the included
approximate equality tests is desired, execute SET-MFT-MODE-ABS,
SET-MFT-MODE-REL, SET-MFT-MODE-AREL, or SET-MFT-MODE-REL0. 
Corresponding to these, the public fvariables MFT-ABS-TOLERANCE,
MFT-REL-TOLERANCE, and MFT-AREL-TOLERANCE contain the values to be used
in comparisons by the words MFT-ABS=, MFT-REL=, and MFT-AREL=. 
MFT-REL0= conditionally executes both MFT-ABS= and MFT-REL=.

Here is an example which uses MFT-REL= :

  SET-MFT-MODE-REL
  mf" 1e-6" MFT-REL-TOLERANCE mf!
  mf{ mf" 1.414214" mf-> mf" 2" mf-sqrt }mf

Note that comparisons in this mode are not symmetric.  Measured values
must be in the mfresults, and reference values in the mfgoals.

The user may supply a different equality test by storing his own xt in
the public variable MFT-TEST=-XT.

Storage of mfinitials, mfresults, and mfgoals includes overflow
checking. The user can change the default storage allocation by defining
one or both of the following CONSTANT's *before* loading this file:

  MAX-MFT-INITIALS  MAX-MFT-RESULTS

The default xt's and tolerances can be changed after loading this file.

Loading mftester.fs does not change BASE.
)
BASE @ DECIMAL

\ *** ERRORS

VARIABLE MFT-ERROR-XT
VARIABLE MFT-ERROR-INDEX   \ set to zero after unequal array size
                           \ comparison, or [index+1] after deepest
                           \ unequal element comparison at index

: MFT-ERROR-DEFAULT  ( c-addr u -- )
(
Display an error message followed by the line that had the error.
)
  TYPE SOURCE TYPE CR ;
' MFT-ERROR-DEFAULT MFT-ERROR-XT !

: MFT-ERROR MFT-ERROR-XT @ EXECUTE ;  \ vectored error reporting

\ *** COMPARISONS

VARIABLE MFT-TEST=-XT

: MFT-TEST=  (mf: x y -- ) ( -- flag )  MFT-TEST=-XT @ EXECUTE ;

\ The sign of these tolerances must be plus.
mf" 0e0"   MFCONSTANT MFT-ABS-TOLERANCE-DEFAULT
mf" 1e-12" MFCONSTANT MFT-REL-TOLERANCE-DEFAULT
MFVARIABLE MFT-ABS-TOLERANCE   MFT-ABS-TOLERANCE-DEFAULT MFT-ABS-TOLERANCE mf!
MFVARIABLE MFT-REL-TOLERANCE   MFT-REL-TOLERANCE-DEFAULT MFT-REL-TOLERANCE mf!

: MFT-DATUM=  (mf: x y -- ) ( -- flag )
(
Leave TRUE if the two mfloats have equivalent internal representations.
Else leave FALSE.
)
  mf= ;

\ |x - y| < eps
: MFT-ABS=  (mf: x y -- ) ( -- flag )
  mf- |mf| MFT-ABS-TOLERANCE mf@ mf<  ;

\ |m - r| < eps * |r|  (note the asymmetry)
: MFT-REL=  (mf: meas ref -- ) ( -- flag )
  mfswap mf-reldiff |mf| MFT-REL-TOLERANCE mf@ mf< ;

: MFT-REL0=  (mf: meas ref -- ) ( -- flag )
  mfdup mf0= IF MFT-ABS= ELSE MFT-REL= THEN ;

: SET-MFT-MODE-EXACT ( -- )  ['] MFT-DATUM= MFT-TEST=-XT ! ;
: SET-MFT-MODE-ABS   ( -- )  ['] MFT-ABS=   MFT-TEST=-XT ! ;
: SET-MFT-MODE-REL   ( -- )  ['] MFT-REL=   MFT-TEST=-XT ! ;
: SET-MFT-MODE-REL0  ( -- )  ['] MFT-REL0=  MFT-TEST=-XT ! ;

SET-MFT-MODE-EXACT

\ *** ARRAYS
(
The first array slot has index zero.  A tick prefix in stack somments
means "address".
)

\ USER CONFIG DEFAULTS
[UNDEFINED] MAX-MFT-INITIALS [IF] 32 CONSTANT MAX-MFT-INITIALS [THEN]
[UNDEFINED] MAX-MFT-RESULTS  [IF] 32 CONSTANT MAX-MFT-RESULTS  [THEN]

\ Array buffers begin with a cell-sized count.
CREATE MFT-INITIALS MAX-MFT-INITIALS 1+ CELLS ALLOT
CREATE MFT-RESULTS  MAX-MFT-RESULTS  1+ CELLS ALLOT
CREATE MFT-GOALS    MAX-MFT-RESULTS  1+ CELLS ALLOT

: 0MFT-A  ( +n 'buf -- )  \ assume n<>0
(
Initialize the array buffer for garbage collection by clearing the count
and n, mfelement addresses.
)
  swap 1+ cells over + swap DO 0 i ! 1 cells +LOOP ;

MAX-MFT-INITIALS MFT-INITIALS 0MFT-A
MAX-MFT-RESULTS  MFT-RESULTS  0MFT-A
MAX-MFT-RESULTS  MFT-GOALS    0MFT-A

\ Overflow checking to be done elsewhere.
: MFT-A!  ( +n 'buf -- ) ( mf: mf_[n-1] ... mf_0 -- )
  2dup ! cell+ swap ( 'elems n) dup
  IF   cells over + swap
       DO i mf! 1 cells +LOOP 
  ELSE 2drop THEN ;

: MFT-A@  ( 'buf -- +n ) ( mf: -- mf_[n-1] ... mf_0 )
  dup @ dup >r ( 'buf n r: n)
  IF   cell+  r@ 1- cells over + 
       DO i mf@ -1 cells +LOOP
  ELSE drop THEN r> ;

[UNDEFINED] CELL- [IF]
  : CELL-  ( addr -- addr-cell )  [ 1 CELLS ] LITERAL - ;
[THEN]

: MFT-A=  ( 'buf1 'buf2 -- 0 true | 0 false | index+1 false )
(
Note that when MFT-TEST=-XT is set to MFT-REL=, the comparison is not
symmetric; so the elements of the measured and reference array buffers
at 'buf1 and 'buf2, respectively, must be fetched to the mfstack in
proper order.  In the stack comments below, "m" and "r" stand for
"measured" and "reference", to keep track in case the comparison xt is
MFT-REL=.

Returns:
         0 true:  equal size and elements
        0 false:  sizes unequal
  index+1 false:  sizes equal, deepest unequal elements at index
)
  2dup @ swap @ <>
  IF 2drop 0 false EXIT THEN          \ unequal size
  dup @ 0= IF 2drop 0 true EXIT THEN  \ equal size and empty
  2>r 0 true 2r>
  ( 0 true 'buf1 'buf2)
  dup @ ( count) >r    \ last index + 1 = count
  r@ cells + swap   r@ cells +   1 r>
  ( 0 true 'r_last 'm_last 1 last+1)
  DO  ( 0 true 'r_index 'm_index) \ i = index+1
    2dup mf@ mf@ ( mf: m_index r_index) MFT-TEST= 0=
    IF  \ unequal elements
      2drop ( 0 true) 2drop i false 0 0 LEAVE
    THEN
    cell- swap cell- swap ( 'r_[index-1] 'm_[index-1])
  -1 +LOOP
  ( [0 true | index+1 false] x1 x2) 2drop ;

\ *** TESTS

: MF{  ( mf: mfinitials -- )
(
Save the mfinitials for restoration by }MF, leaving the mfstack empty. 
Clear mfresults and mfgoals storage.
)
  mfdepth dup MAX-MFT-INITIALS > ABORT" TOO MANY MFINITIALS"
  MFT-INITIALS MFT-A!   0 MFT-RESULTS !   0 MFT-GOALS ! ;

: MF->  ( mf: mfesults -- )
(
Save the mfresults, leaving the mfstack empty.
)
  mfdepth dup MAX-MFT-RESULTS > ABORT" TOO MANY MFRESULTS"
  MFT-RESULTS MFT-A! ;

: }MF  ( mf: mfgoals -- mfinitials )
(
Save the mfgoals, compare them with the saved mfresults, and restore the
mfinitials.  If the number of mfresults is not the same as the number of
mfgoals, store zero in MFT-ERROR-INDEX and execute the MFT-ERROR vector. 
Else, if the numbers are the same but there is a deepest comparison
failure at the index-th element, store index+1 in MFT-ERROR-INDEX and
execute the MFT-ERROR vector. 
)
  mfdepth dup MAX-MFT-RESULTS > ABORT" TOO MANY MFGOALS"
  MFT-GOALS MFT-A!
  MFT-RESULTS MFT-GOALS MFT-A= ( 0 true | 0 false | index+1 false)
  IF
    ( 0) drop
  ELSE
    dup MFT-ERROR-INDEX !
    IF S" INCORRECT MFRESULT: " ELSE S" WRONG NUMBER OF MFRESULTS: " THEN
    MFT-ERROR
  THEN
  MFT-INITIALS MFT-A@ ( len) drop ;

BASE !
\ *** END of mftester.fs
